!+
! Module: makemask_mod
!
! Authors:  Steven L. Snowden - Goddard Space Flight Center (NASA) Original code.
!           Brendan Perry -- f95 comprehensive rewrite for SAS compliance Spring 2020
!
! Description: Extended Source Analysis Software
!
! Comments:
!
! $Id: makemask_mod.f90,v 1.2 2009/02/20 13:27:08 aibarra Exp $
!-
!  Version 5.0  9 November 2012
!
!  Called by cheese in mode=2 to create a cheese mask
!
!  MAKE-MASK is called by the PERL script CHEESE to create a
!  swiss cheese mask.   bhp: ALSO called by CHEESE-BANDS!
!
!  Parameters:
!
!    imagefile  (INPUT event image for the exposure.)
!    maskfile   (INPUT exposure mask.)
!    regionfile (INPUT filtered source region list.)
!    cheesefile (OUTPUT cheese mask image in sky coordinates)
!
!    Example: 
!              makemask imagefile=mos1S001-fovim.fits \
!              maskfile=mos1S001-maskimt.fits \
!              regionfile=mos1S001-bkgregt-sky.fits \
!              cheesefile=mos1S001-cheeset.fits
! input image, mos1S001-obj-im.fits, is now mos1S001-fovimt.fits
!
module makemask_mod

  use dal
  use oal
  use param
  use errorHandling
  use gofio_mod
  use gofutils_mod
  use coordinate   ! for doing coordinate conversions (ra/dec->x,y)

  implicit none

  private

  character(len=*), parameter :: &
       inimagepar  = "imagefile" , &
       inmaskpar   = "maskfile" , &
       inregionpar = "regionfile", &
       outmaskpar  = "cheesefile"

  public makemask

contains

  !---------------------------------------------------------------------
  subroutine makemask
    !---------------------------------------------------------------------

    !
    ! SAS OAL datatypes:
    !
    type(DataSetT) :: inimage_S, inmask_S, inregion_S, outcheese_S
    type(ArrayT)   :: inmask_A, outcheese_A, inimage_A
    type(TableT)   :: inregion_T, outregion_T
    type(ColumnT)  :: region_C
    type(SpacecraftAttitudeType) :: scAttitude, instrumAttitude
    !
    character(len=16), allocatable, dimension(:) :: shape
    character(len=256) :: inimage_F, inmask_F, outcheese_F, inregion_F
    character(len=5) ::  ctyp, expidstr
    character(len=8) ::  ctypeI(2), ctypeM(2), regextnam = "REGION"
    character(len=10) ::  cunitI(2), cunitM(2), radecsysI, radecsysM
    character(len=20) :: instrume
    ! always 900x900???!!!
    integer :: xsizeM=900, ysizeM=900 ! bhp: adjust if necessary (check this!!!)
    integer :: i, j, k, icol, ir, irad, irot, ishape, ix, ixrc, iy, &
         iyrc, j1, j2, jj, jj1, jj2, lun, inregrows, status, nvect
    integer(int16) :: inmask_A_Xsize, inmask_A_Ysize
    integer(int32) :: inmask_A_type, instrumentId
    integer(int32), dimension(:), pointer :: inmask_A_shape
    integer(int32), dimension(:,:), pointer :: outcheese_P, inmask_P
    integer, allocatable :: mask(:,:)    
    logical :: exists=.FALSE., warndie
    real(kind=single) :: rad, rad1, rad2, rotang, rr, & 
         xrc, xrefpix, xp, xpixdelt, xx, &
         yrc, yrefpix, yp, ypixdelt, yy
    real(kind=single), dimension(:,:), pointer :: sha, x_P, y_P, r_P, rotang_P
    real(kind=double) :: cdeltI(2), crpixI(2), crvalI(2), crotI, crotM, &
         ra_pntM, dec_pntM, ra_pntI, dec_pntI, pa_pntI, pa_pntM, &
         xrefpixI, yrefpixI, xpixdeltI, ypixdeltI, xrefpixM, yrefpixM, &
         xpixdeltM, ypixdeltM
    real(kind=double) :: cdeltM(2), crpixM(2), crvalM(2), crotaI(2), crotaM(2)
    real(kind=double) :: expoI, expoM, expStartTimeStamp
    real(kind=single), allocatable :: r(:,:), rot(:), x(:), y(:)
    inregion_F = ""
    inimage_F = ""
    inmask_F = ""
    outcheese_F = ""
    ! bhp: read the parameters from the parameter file:
    inimage_F   = stringparameter(inimagepar)
    inmask_F    = stringparameter(inmaskpar)
    outcheese_F = stringparameter(outmaskpar)
    inregion_F  = stringparameter(inregionpar)
    !
    write(errstr,*) "Event Image: ",trim(inimage_F)
    call message(AppMsg,Always)
    write(errstr,*) "Input Exp Mask file: ",trim(inmask_F)
    call message(AppMsg,Always)
    write(errstr,*) "Output Cheese Mask file: ",trim(outcheese_F)
    call message(AppMsg,Always)
    write(errstr,*) "Input Region file: ",trim(inregion_F)
    call message(AppMsg,Always)
    !
    ! bhp: check and open the input image file (fatal if not found)
    !
    write(errstr,*) "MAKEMASK: Open/read input image: ",trim(inimage_F)
    call message(AppMsg,Always)
    status = 0
    warndie = .TRUE. ! will die if there's a problem
    call check_open_set(trim(inimage_F),inimage_S,1,status,warndie)
    !
    ! bhp: get the cal info for the coordinate transforms later
    ! in ssclib/cal_aux. Sets CAL instrument state. ???!!!
    !call getCalInfo(inregion_S, instrumentId, expStartTimeStamp, &
    !     scAttitude, instrumAttitude) 
    !if ((instrumentId .ne. EMOS1) &
    !     .and. (instrumentId .ne. EMOS2) &
    !     .and. (instrumentId .ne. EPN)) then
    !   call error('badInstrument', 'Instrument '&
    !        //trim(instId(instrumentId))//' not supported.')
    !end if
    !
    ! BHP: only need certain WCS keywords found in primary header file
    ! 
    write(errstr,*) "Reading keywords from input image file."
    call message(AppMsg,Always)
    ! these are in gofio in goflib
    call coord_keys (inimage_S, ra_pntI, dec_pntI, pa_pntI, crvalI, cdeltI, &
         cunitI, ctypeI, crpixI, crotaI, radecsysI, warndie)
    call more_coord_keys (inimage_S,xrefpixI,yrefpixI,xpixdeltI,ypixdeltI,warndie)
    !
    ! get image exposure
    !
    inimage_A = array(inimage_S,"PRIMARY", Read)
    if (hasAttribute(inimage_S,"EXPOSURE")) then
       expoI = real64Attribute(inimage_S,"EXPOSURE")
    elseif (hasAttribute(inimage_A,"EXPOSURE")) then
       expoI = real64Attribute(inimage_A,"EXPOSURE")
    else
       call error("missingAttribute",&
            "Keyword EXPOSURE is missing from input image file")
    end if
    !
    ! Close the input image file
    !
    write(errstr,*) "Closing input image file."
    call message(AppMsg,Always)
    call release(inimage_S) 
    !
    ! BHP: check and open the input mask file (fatal if not found)
    !
    write(errstr,*) "MAKEMASK: Processing input mask: ",trim(inmask_F)
    call message(AppMsg,Always)
    status = 0
    warndie = .TRUE.
    call check_open_set(trim(inmask_F),inmask_S,1,status,warndie)
    !
    ! BHP: only need certain WCS keywords ('attributes'):
    !      ONLY found in primary header file!!!
    write(errstr,*) "Reading keywords from input mask file."
    call message(AppMsg,Always)
    call coord_keys (inmask_S, ra_pntM, dec_pntM, pa_pntM, crvalM, cdeltM, &
         cunitM, ctypeM, crpixM, crotaM, radecsysM, warndie)
    call more_coord_keys (inmask_S,xrefpixM,yrefpixM,xpixdeltM,ypixdeltM,warndie)
    write(errstr,*) "MAKEMASK: coordkeys = ",xrefpixM,yrefpixM,xpixdeltM,ypixdeltM
    call message(AppMsg,Noisy)
    !
    ! get mask exposure
    !
    inmask_A = array(inmask_S,"MASK", Read)
    if (hasAttribute(inmask_S,"EXPOSURE")) then
       expoM = real64Attribute(inmask_S,"EXPOSURE")
    elseif (hasAttribute(inmask_A,"EXPOSURE")) then
       expoM = real64Attribute(inmask_A,"EXPOSURE")
    else
       call error("missingAttribute",&
            "Keyword EXPOSURE is missing from input mask file")
    end if    ! compare I and M values above (sanity check)???!!!
    !
    ! BHP: read input mask, get shape 
    !      ???!!! should be 900,900 I*4, right???!!!
    !
    inmask_A = array(inmask_S,"MASK", Read)
    inmask_A_type = dataType(inmask_A)
    inmask_A_shape => dimensions(inmask_A) ! vector pointer
    inmask_A_Xsize = inmask_A_shape(0) 
    inmask_A_Ysize = inmask_A_shape(1) 
    !
    ! BHP: see if mask naxis1 and naxis2 are 900x900 again always???!!!
    !      make sure arrays align!!!!
    if ( (inmask_A_Xsize .ne. 900) .or. &
         (inmask_A_Ysize .ne. 900) ) then 
       call warning("Input mask shape is not 900x900!" )
    end if
    inmask_P => int32Array2Data(inmask_A) ! pointer to mask array
    !
    ! ss:  Close the input mask file
    !
    write(errstr,*) "Closing input mask file."
    call message(AppMsg,Always)
    call release(inmask_S) 
    !
    ! ss: Print the details of the template image
    !
    write(errstr,*) "Template (inmask) File: ",trim(inmask_F)
    call message(AppMsg,Always)
    write(errstr,*) "Template (inmask) image coordinate system: ",radecsysM
    call message(AppMsg,Always)
    write(errstr,*) "Template (inmask) image coordinate ID: ", &
         ctypeM(1), "  ", ctypeM(2)
    call message(AppMsg,Always)
    write(errstr,*) "Template image central coordinates: ", &
         crvalM(1), "  ",crvalM(2)
    call message(AppMsg,Always)
    write(errstr,*) "Template image reference pixel: ", &
         crpixM(1), "  ",crpixM(2)
    call message(AppMsg,Always)
    write(errstr,*) "Template image pixel sizes: ", &
         cdeltM(1), "  ",cdeltM(2)
    call message(AppMsg,Always)
    !
    write(errstr,*) "MAKEMASK: Processing region file: ",trim(inregion_F)
    call message(AppMsg,Always)
    status = 0
    warndie = .TRUE. ! fatal if file not ok
    call check_open_set(trim(inregion_F),inregion_S,1,status,warndie)
    if ( hasBlock(inregion_S,trim(regextnam)) ) then ! REGION
       inregion_T = table(inregion_S,regextnam)
    else
       write(errstr,*)"Region file exists, but is not EXTNAM =  ",&
            trim(regextnam)
       call error("NoREGIONext",trim(errstr))
    end if
    inregrows = numberOfRows(inregion_T)
    if (inregrows .le. 0) then
       call error("NoREGIONrows",&
            "Region file extension has 0 rows")
    end if
    !
    ! ss: now read the data:
    !
    write(errstr,*) "Reading data from region file..."
    call message(AppMsg,Always)
    region_C = column(inregion_T,"SHAPE", Read)  !typically 16A
    allocate(shape(inregrows))
    do j=1,inregrows
       shape(j) = stringCell(region_C,j-1)
    end do
    call release(region_C)
    region_C = column(inregion_T,"X", Read)      !typically 4E vector
    x_P => real32VectorData(region_C)
    nvect = numberOfdimensions(region_C)
    call release(region_C)
    region_C = column(inregion_T,"Y", Read)      !typically 4E vector
    y_P => real32VectorData(region_C)
    call release(region_C)
    region_C = column(inregion_T,"R", Read)      !typically 4E vector
    r_P => real32VectorData(region_C)
    call release(region_C)
    region_C = column(inregion_T,"ROTANG", Read) !typically 4E vector
    rotang_P => real32VectorData(region_C)
    call release(region_C)
    call release(inregion_T)
    !
    write(errstr,*) "Number of excluded sources: ",inregrows
    call message(AppMsg,Always)
    
    do i = 0, inregrows-1
       write(errstr,*) 'SHAPE = ',trim(adjustl(shape(i+1)))
       call message(AppMsg,Verbose)
       if(trim(adjustl(shape(i+1))) .eq. '!CIRCLE') then
          write(errstr,*) 'PrecheckC - ',i,crpixM(1),xpixdeltM,x_P(0,i),xrefpixM,cdeltM(1)
          call message(AppMsg,Verbose)
          write(errstr,*) 'PrecheckC - ',i,crpixM(2),ypixdeltM,y_P(0,i),xrefpixM,cdeltM(2)
          call message(AppMsg,Verbose)
          xrc = crpixM(1) + &
               xpixdeltM*(x_P(0,i) - xrefpixM)/cdeltM(1)
          yrc = crpixM(2) + &
               ypixdeltM*(y_P(0,i) - yrefpixM)/cdeltM(2)
          rad = r_P(0,i)*ypixdeltM/cdeltM(2)
          ixrc = xrc
          iyrc = yrc
          irad = rad
          !  Check to see whether the radius has exploded
          write(errstr,*)'OverC - ',i,ixrc,iyrc,irad
          call message(AppMsg,Verbose)
          !
          if(irad .gt. 0) then
             j1 = ixrc - irad
             j2 = ixrc + irad
             if(j1 .lt. 1) j1 = 1
             if(j2 .gt. 900) j2 = 900
             do j=j1,j2
                jj1 = iyrc - irad
                jj2 = iyrc + irad
                if(jj1 .lt. 1) jj1 = 1
                if(jj2 .gt. 900) jj2 = 900
                do jj=jj1,jj2
                   rr = sqrt((j-xrc)**2 + (jj-yrc)**2)
                   if (rr .lt. rad) then
                      inmask_P(j,jj) = 0.
                   endif
                enddo
             enddo
          endif
       elseif(trim(adjustl(shape(i+1))) .eq. '!ELLIPSE') then
          ! change to Verbose???!!!
          write(errstr,*) 'PrecheckE - ',i,crpixM(1),xpixdeltM,x_P(0,i),xrefpixM,cdeltM(1)
          call message(AppMsg,Verbose)
          write(errstr,*) 'PrecheckE - ',i,crpixM(2),ypixdeltM,y_P(0,i),xrefpixM,cdeltM(2)
          call message(AppMsg,Verbose)          
          xrc = crpixM(1) + &
               xpixdeltM*(x_P(0,i) - xrefpixM)/cdeltM(1)
          yrc = crpixM(2) + &
               ypixdeltM*(y_P(0,i) - yrefpixM)/cdeltM(2)
          rad1 = r_P(0,i)*ypixdeltM/cdeltM(2)
          rad2 = r_P(1,i)*ypixdeltM/cdeltM(2)
          ixrc = xrc
          iyrc = yrc
          irad = rad1
          if(rad1 .lt. rad2) irad = rad2
          rotang = -rotang_P(0,i)*3.14159263/180.0
          write(errstr,*) 'OverE - ',i,ixrc,iyrc,irad,rotang
          call message(AppMsg,Verbose) 
          !
          !  Check to see whether the radius has exploded
          !
          rad1 = rad1*rad1
          rad2 = rad2*rad2
          if(irad .gt. 0) then
             j1 = ixrc - irad
             j2 = ixrc + irad
             if(j1 .lt. 1) j1 = 1
             if(j2 .gt. 900) j2 = 900
             jj1 = iyrc - irad
             jj2 = iyrc + irad
             if(jj1 .lt. 1) jj1 = 1
             if(jj2 .gt. 900) jj2 = 900
             do j=j1,j2
                do jj=jj1,jj2
                   xx = j - ixrc
                   yy = jj - iyrc
                   xp = xx*cos(rotang) - yy*sin(rotang)
                   yp = yy*cos(rotang) + xx*sin(rotang)
                   rr = xp*xp/rad1 + yp*yp/rad2
                   if (rr .lt. 1.0) then
                      inmask_P(j,jj) = 0.
                   end if
                enddo
             enddo
          endif
          ! BHP: put something here to catch 'neither CIRCLE nor ELLIPSE'???!!!
       endif
    enddo
    !
    ! BHP: open output FITS file, set up primary array:
    !
    write(errstr,*) "MAKEMASK: opening output cheesefile for write"
    call message(AppMsg,Always)
    outcheese_S = dataSet(trim(outcheese_F), Create)
    outcheese_A = addArray(outcheese_S, "PRIMARY", INTEGER32,&
         inmask_A_shape) 
    outcheese_P => int32Array2Data(outcheese_A)
    !
    ! BHP: fill mask with mask() array
    !
    do j=0,inmask_A_Xsize-1    ! should 'always' be 900 ???!!!
       do k=0,inmask_A_Ysize-1 ! should 'always' be 900 ???!!!
          outcheese_P(j,k) = int(inmask_P(j,k)) ! check this ??!!!
       end do
    end do
    !
    !  ss: Write out the new mask (or ok to clone???!!!)
    !
    write(errstr,*) "MAKEMASK: copying keywords to cheese mask: "
    call message(AppMsg,Always)
    !  ss: Write some useful header keywords
    ! BHP: call copyAllAttributes???!!!
    call setAttribute(outcheese_S,'CONTENT','IMAGE','Mask Image')
    call setAttribute(outcheese_S,'TELESCOP','XMM',&
         'Telescope (mission) name')
    call setAttribute(outcheese_S,'INSTRUME','EPIC','Instrument name')
    call setAttribute(outcheese_S,'OBS_MODE','POINTING', &
         'Obs Mode: POINTING,SLEW, OR SCAN')
    call setAttribute(outcheese_S,'RADECSYS','FK5', &
         'Equatorial system reference')
    call setAttribute(outcheese_S,'EQUINOX',2000,'Equinox')
    call setAttribute(outcheese_S,'EXPOSURE',expoI,'Exposure Time','s')
    ! Removed SAS-21 because it's a reserved keyword
    !    call setAttribute(outcheese_S,'BSCALE',1.0,'Scale',' ')     ! BHP: are these
    call setAttribute(outcheese_S,'BZERO',0.0,'Zero Point',' ') ! BHP: even used???!!!
    call setAttribute(outcheese_S,'BUNIT','counts','Units of data')
    call setAttribute(outcheese_S,'CRPIX1',crpixI(1),'Reference pixel')
    call setAttribute(outcheese_S,'CTYPE1',ctypeI(1),'Projection')
    call setAttribute(outcheese_S,'CRVAL1',crvalI(1),'Central longitude')
    call setAttribute(outcheese_S,'CDELT1',cdeltI(1),'Pixel size')
    call setAttribute(outcheese_S,'CUNIT1','deg','Units of coordinate')
    call setAttribute(outcheese_S,'CRPIX2',crpixI(2),'Reference pixel')
    call setAttribute(outcheese_S,'CTYPE2',ctypeI(2),'Projection')
    call setAttribute(outcheese_S,'CRVAL2',crvalI(2),'Central latitude')
    call setAttribute(outcheese_S,'CDELT2',cdeltI(2),'Pixel size')
    call setAttribute(outcheese_S,'CUNIT2','deg','Units of coordinate')

    !
    ! BHP: should change these to reflect SAS-correct code at some point???!!!
    !
    call addComment(outcheese_S, &
         'This is a source mask (Swiss cheese mask) created by the ')
    call addComment(outcheese_S, &
         'program makemask using a merged source list output ')
    call addComment(outcheese_S,'from the SAS task cheese')

    !
    !  ss: Print the details of the output image
    !
    write(errstr,*)'Output image coordinate system: ',trim(adjustl(radecsysI))
    call message(AppMsg,Always)
    write(errstr,*)'Output image coordinate ID: ',   &
         trim(adjustl(ctypeM(1))), '  ', trim(adjustl(ctypeM(2)))
    call message(AppMsg,Always)
    write(errstr,*)'Output image central coordinates: ',  & 
         crvalM(1), '  ',crvalM(2)
    call message(AppMsg,Always)
    write(errstr,*)'Output image reference pixel: ',   &
         crpixM(1), '  ',crpixM(2)
    call message(AppMsg,Always)

    write(errstr,*)"Wrote TOTAL cheese mask file: ",trim(outcheese_F)
    call message(AppMsg,Always)
    !
    ! BHP: this closes the total band cheese mask file:
    !
    call release(outcheese_A)
    call release(outcheese_S) 

    return

  end subroutine makemask

end module makemask_mod
